"""
browsr command line interface
"""

from typing import Optional

import click
import rich_click

from browsr._base import (
    TextualAppContext,
)
from browsr._version import __application__, __version__
from browsr.browsr import Browsr

rich_click.rich_click.MAX_WIDTH = 100
rich_click.rich_click.STYLE_OPTION = "bold green"
rich_click.rich_click.STYLE_SWITCH = "bold blue"
rich_click.rich_click.STYLE_METAVAR = "bold red"
rich_click.rich_click.STYLE_HELPTEXT_FIRST_LINE = "bold blue"
rich_click.rich_click.STYLE_HELPTEXT = ""
rich_click.rich_click.STYLE_HEADER_TEXT = "bold green"
rich_click.rich_click.STYLE_OPTION_DEFAULT = "bold yellow"
rich_click.rich_click.STYLE_OPTION_HELP = ""
rich_click.rich_click.STYLE_ERRORS_SUGGESTION = "bold red"
rich_click.rich_click.STYLE_OPTIONS_TABLE_BOX = "SIMPLE_HEAVY"
rich_click.rich_click.STYLE_COMMANDS_TABLE_BOX = "SIMPLE_HEAVY"


@click.command(name="browsr", cls=rich_click.rich_command.RichCommand)
@click.argument("path", default=None, required=False, metavar="PATH")
@click.option(
    "-m",
    "--max-file-size",
    default=20,
    type=int,
    help="Maximum file size in MB for the application to open",
)
@click.version_option(version=__version__, prog_name=__application__)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable extra debugging output",
    type=click.BOOL,
)
def browsr(
    path: Optional[str],
    debug: bool,
    max_file_size: int,
) -> None:
    """
    browsr 🗂️  a pleasant file explorer in your terminal

    Navigate through directories and peek at files whether they're hosted locally,
    over SSH, in GitHub, AWS S3, Google Cloud Storage, or Azure Blob Storage. View code files
    with syntax highlighting, format JSON files, render images, convert data files to navigable
    datatables, and more.

    \f

    ![browsr](https://raw.githubusercontent.com/juftin/browsr/main/docs/_static/screenshot_utils.png)

    ## Installation

    It's recommended to install **`browsr`** via [pipx](https://pypa.github.io/pipx/)
    with **`all`** optional dependencies, this enables **`browsr`** to access
    remote cloud storage buckets and open parquet files.

    ```shell
    pipx install "browsr[all]"
    ```

    ## Usage Examples

    ### Local

    #### Browse your current working directory

    ```shell
    browsr
    ```

    #### Browse a local directory

    ```shell
    browsr/path/to/directory
    ```

    ### Cloud Storage

    #### Browse an S3 bucket

    ```shell
    browsr s3://bucket-name
    ```

    #### Browse a GCS bucket

    ```shell
    browsr gs://bucket-name
    ```

    #### Browse Azure Services

    ```shell
    browsr adl://bucket-name
    browsr az://bucket-name
    ```

    ### GitHub

    #### Browse a GitHub repository

    ```shell
    browsr github://juftin:browsr
    ```

    #### Browse a GitHub Repository Branch

    ```shell
    browsr github://juftin:browsr@main
    ```

    #### Browse a Private GitHub Repository

    ```shell
    export GITHUB_TOKEN="ghp_1234567890"
    browsr github://juftin:browsr-private@main
    ```

    #### Browse a GitHub Repository Subdirectory

    ```shell
    browsr github://juftin:browsr@main/tests
    ```

    #### Browse a GitHub URL

    ```shell
    browsr https://github.com/juftin/browsr
    ```

    #### Browse a Filesystem over SSH

    ```
    browsr ssh://user@host:22
    ```

    #### Browse a SFTP Server

    ```
    browsr sftp://user@host:22/path/to/directory
    ```

    ## Key Bindings
    - **`Q`** - Quit the application
    - **`F`** - Toggle the file tree sidebar
    - **`T`** - Toggle the rich theme for code formatting
    - **`N`** - Toggle line numbers for code formatting
    - **`D`** - Toggle dark mode for the application
    - **`X`** - Download the file from cloud storage
    """
    config = TextualAppContext(file_path=path, debug=debug, max_file_size=max_file_size)
    app = Browsr(config_object=config)
    app.run()


if __name__ == "__main__":
    browsr()



"""
Pytest Fixtures Shared Across all Unit Tests
"""

import pathlib
from typing import Any, Dict, List

import pytest
from click.testing import CliRunner
from textual_universal_directorytree import GitHubPath


@pytest.fixture
def runner() -> CliRunner:
    """
    Return a CliRunner object
    """
    return CliRunner()


@pytest.fixture
def repo_dir() -> pathlib.Path:
    """
    Return the path to the repository root
    """
    return pathlib.Path(__file__).parent.parent.resolve()


@pytest.fixture
def screenshot_dir(repo_dir: pathlib.Path) -> pathlib.Path:
    """
    Return the path to the screenshot directory
    """
    return repo_dir / "tests" / "screenshots"


@pytest.fixture
def github_release_path() -> GitHubPath:
    """
    Return the path to the Github Release
    """
    release = "v1.6.0"
    uri = f"github://juftin:browsr@{release}"
    return GitHubPath(uri)


@pytest.fixture(scope="module")
def vcr_config() -> Dict[str, List[Any]]:
    """
    VCR Cassette Privacy Enforcer

    This fixture ensures the API Credentials are obfuscated

    Returns
    -------
    Dict[str, list]:
    """
    return {
        "filter_headers": [("authorization", "XXXXXXXXXX")],
        "filter_query_parameters": [("user", "XXXXXXXXXX"), ("token", "XXXXXXXXXX")],
    }


cassette = pytest.mark.vcr(scope="module")



"""
Config / Context Tests
"""
import pathlib
from dataclasses import is_dataclass

from textual_universal_directorytree import GitHubPath

from browsr._base import TextualAppContext
from browsr._config import favorite_themes
from tests.conftest import cassette


def test_favorite_themes() -> None:
    """
    Test that favorite_themes is a list of strings
    """
    assert isinstance(favorite_themes, list)
    for theme in favorite_themes:
        assert isinstance(theme, str)


def test_textual_app_context() -> None:
    """
    Test that TextualAppContext is a dataclass
    """
    assert is_dataclass(TextualAppContext)


def test_textual_app_context_path() -> None:
    """
    Test that default TextualAppContext.path is CWD
    """
    context = TextualAppContext()
    assert isinstance(context.path, pathlib.Path)
    assert context.path == pathlib.Path.cwd().resolve()


@cassette
def test_textual_app_context_path_github() -> None:
    """
    Test GitHub URL Parsing
    """
    github_strings = [
        "https://github.com/juftin/browsr",
        "https://github.com/juftin/browsr.git",
        "github.com/juftin/browsr",
        "www.github.com/juftin/browsr",
        "https://www.github.com/juftin/browsr",
        "github://juftin:browsr",
        "github://juftin:browsr@main",
    ]
    for _github_string in github_strings:
        context = TextualAppContext(file_path=_github_string)
        handled_github_url = context.path
        expected_file_path = "github://juftin:browsr@main/"
        assert handled_github_url == GitHubPath(expected_file_path)
        assert str(handled_github_url) == expected_file_path



"""
browsr configuration file
"""

from os import getenv
from typing import List

favorite_themes: List[str] = [
    "monokai",
    "material",
    "dracula",
    "solarized-light",
    "one-dark",
    "solarized-dark",
    "emacs",
    "vim",
    "github-dark",
    "native",
    "paraiso-dark",
]
rich_default_theme = getenv("RICH_THEME", None)

if rich_default_theme in favorite_themes:
    assert isinstance(rich_default_theme, str)
    favorite_themes.remove(rich_default_theme)
if rich_default_theme is not None:
    assert isinstance(rich_default_theme, str)
    favorite_themes.insert(0, rich_default_theme)

image_file_extensions = [
    ".bmp",
    ".dib",
    ".eps",
    ".ps",
    ".gif",
    ".icns",
    ".ico",
    ".cur",
    ".im",
    ".im.gz",
    ".im.bz2",
    ".jpg",
    ".jpe",
    ".jpeg",
    ".jfif",
    ".msp",
    ".pcx",
    ".png",
    ".ppm",
    ".pbm",
    ".pgm",
    ".sgi",
    ".rgb",
    ".bw",
    ".spi",
    ".tif",
    ".tiff",
    ".webp",
    ".xbm",
    ".xv",
    ".pdf",
]



"""
Code Browsr Utility Functions
"""

import datetime
import os
import pathlib
from dataclasses import dataclass
from typing import Any, BinaryIO, Dict, Optional, Union

import fitz  # type: ignore[import]
import rich_pixels
from fitz import Pixmap
from PIL import Image
from rich_pixels import Pixels
from textual_universal_directorytree import is_remote_path


def _open_pdf_as_image(buf: BinaryIO) -> Image.Image:
    """
    Open a PDF file and return a PIL.Image object
    """
    doc = fitz.open(stream=buf.read(), filetype="pdf")
    pix: Pixmap = doc[0].get_pixmap()
    if pix.colorspace is None:
        mode = "L"
    elif pix.colorspace.n == 1:
        mode = "L" if pix.alpha == 0 else "LA"
    elif pix.colorspace.n == 3:  # noqa: PLR2004
        mode = "RGB" if pix.alpha == 0 else "RGBA"
    else:
        mode = "CMYK"
    return Image.frombytes(size=(pix.width, pix.height), data=pix.samples, mode=mode)


def open_image(document: pathlib.Path, screen_width: float) -> Pixels:
    """
    Open an image file and return a rich_pixels.Pixels object
    """
    with document.open("rb") as buf:
        if document.suffix.lower() == ".pdf":
            image = _open_pdf_as_image(buf=buf)
        else:
            image = Image.open(buf)
        image_width = image.width
        image_height = image.height
        size_ratio = image_width / screen_width
        new_width = min(int(image_width / size_ratio), image_width)
        new_height = min(int(image_height / size_ratio), image_height)
        resized = image.resize((new_width, new_height))
        return rich_pixels.Pixels.from_image(resized)


@dataclass
class FileInfo:
    """
    File Information Object
    """

    file: pathlib.Path
    size: int
    last_modified: Optional[datetime.datetime]
    stat: Union[Dict[str, Any], os.stat_result]
    is_local: bool
    is_file: bool
    owner: str
    group: str
    is_cloudpath: bool


def get_file_info(file_path: pathlib.Path) -> FileInfo:
    """
    Get File Information, Regardless of the FileSystem
    """
    try:
        stat: Union[Dict[str, Any], os.stat_result] = file_path.stat()
        is_file = file_path.is_file()
    except PermissionError:
        stat = {"size": 0}
        is_file = True
    is_cloudpath = is_remote_path(file_path)
    if isinstance(stat, dict):
        lower_dict = {key.lower(): value for key, value in stat.items()}
        file_size = lower_dict["size"]
        modified_keys = ["lastmodified", "updated", "mtime"]
        last_modified = None
        for modified_key in modified_keys:
            if modified_key in lower_dict:
                last_modified = lower_dict[modified_key]
                break
        if isinstance(last_modified, str):
            last_modified = datetime.datetime.fromisoformat(last_modified[:-1])
        return FileInfo(
            file=file_path,
            size=file_size,
            last_modified=last_modified,
            stat=stat,
            is_local=False,
            is_file=is_file,
            owner="",
            group="",
            is_cloudpath=is_cloudpath,
        )
    else:
        last_modified = datetime.datetime.fromtimestamp(stat.st_mtime)
        try:
            owner = file_path.owner()
            group = file_path.group()
        except NotImplementedError:
            owner = ""
            group = ""
        return FileInfo(
            file=file_path,
            size=stat.st_size,
            last_modified=last_modified,
            stat=stat,
            is_local=True,
            is_file=is_file,
            owner=owner,
            group=group,
            is_cloudpath=is_cloudpath,
        )


def handle_duplicate_filenames(file_path: pathlib.Path) -> pathlib.Path:
    """
    Handle Duplicate Filenames

    Duplicate filenames are handled by appending a number to the filename
    in the form of "filename (1).ext", "filename (2).ext", etc.
    """
    if not file_path.exists():
        return file_path
    else:
        i = 1
        while True:
            new_file_stem = f"{file_path.stem} ({i})"
            new_file_path = file_path.with_stem(new_file_stem)
            if not new_file_path.exists():
                return new_file_path
            i += 1


def handle_github_url(url: str) -> str:
    """
    Handle GitHub URLs

    GitHub URLs are handled by converting them to the raw URL.
    """
    try:
        import requests
    except ImportError as e:
        raise ImportError(
            "The requests library is required to browse GitHub files. "
            "Install browsr with the `remote` extra to install requests."
        ) from e

    gitub_prefix = "github://"
    if gitub_prefix in url and "@" not in url:
        _, user_password = url.split("github://")
        org, repo_str = user_password.split(":")
        repo, *args = repo_str.split("/")
    elif gitub_prefix in url and "@" in url:
        return url
    elif "github.com" in url.lower():
        _, org, repo, *args = url.split("/")
    else:
        raise ValueError(f"Invalid GitHub URL: {url}")
    token = os.getenv("GITHUB_TOKEN")
    auth = {"auth": ("Bearer", token)} if token is not None else {}
    resp = requests.get(
        f"https://api.github.com/repos/{org}/{repo}",
        headers={"Accept": "application/vnd.github.v3+json"},
        **auth,  # type: ignore[arg-type]
    )
    resp.raise_for_status()
    default_branch = resp.json()["default_branch"]
    arg_str = "/".join(args)
    github_uri = f"{gitub_prefix}{org}:{repo}@{default_branch}/{arg_str}".rstrip("/")
    return github_uri


class ArchiveFileError(Exception):
    """
    Archive File Error
    """


def render_file_to_string(file_info: FileInfo) -> str:
    """
    Render File to String

    Parameters
    ----------
    file_info : FileInfo
        The file to render.

    Returns
    -------
    str
        The rendered file as a string.
    """
    try:
        return file_info.file.read_text(encoding="utf-8")
    except UnicodeDecodeError as e:
        if file_info.file.suffix.lower() in [".tar", ".gz", ".zip", ".tgz"]:
            raise ArchiveFileError(
                f"Cannot render archive file {file_info.file}."
            ) from e
        else:
            raise e



"""
Extension Classes
"""

from __future__ import annotations

import math
import pathlib
from dataclasses import dataclass
from textwrap import dedent
from typing import Any, Dict, Optional, Union

import numpy as np
import upath
from pandas import DataFrame
from rich import traceback
from rich.console import RenderableType
from rich.markdown import Markdown
from rich.text import Text
from textual import on
from textual.app import App, ComposeResult
from textual.containers import Container
from textual.reactive import reactive, var
from textual.widget import Widget
from textual.widgets import Button, DataTable, Static

from browsr._config import favorite_themes
from browsr._utils import FileInfo, handle_github_url


@dataclass
class TextualAppContext:
    """
    App Context Object
    """

    file_path: Optional[str] = None
    config: Optional[Dict[str, Any]] = None
    debug: bool = False
    max_file_size: int = 20

    @property
    def path(self) -> pathlib.Path:
        """
        Resolve `file_path` to a upath.UPath object
        """
        if "github" in str(self.file_path).lower():
            file_path = str(self.file_path)
            file_path = file_path.lstrip("https://")
            file_path = file_path.lstrip("http://")
            file_path = file_path.lstrip("www.")
            if file_path.endswith(".git"):
                file_path = file_path[:-4]
            file_path = handle_github_url(url=str(file_path))
            self.file_path = file_path
        if str(self.file_path).endswith("/"):
            self.file_path = str(self.file_path)[:-1]
        return (
            upath.UPath(self.file_path).resolve()
            if self.file_path
            else pathlib.Path.cwd().resolve()
        )


class BrowsrTextualApp(App[str]):
    """
    textual.app.App Extension
    """

    show_tree = var(True)
    theme_index = var(0)
    linenos = var(False)
    rich_themes = favorite_themes
    selected_file_path: Union[upath.UPath, pathlib.Path, None, var[None]] = var(None)
    force_show_tree = var(False)
    hidden_table_view = var(False)

    def __init__(
        self,
        config_object: Optional[TextualAppContext] = None,
    ):
        """
        Like the textual.app.App class, but with an extra config_object property

        Parameters
        ----------
        config_object: Optional[TextualAppContext]
            A configuration object. This is an optional python object,
            like a dictionary to pass into an application
        """
        super().__init__()
        self.config_object = config_object
        traceback.install(show_locals=True)

    @staticmethod
    def df_to_table(
        pandas_dataframe: DataFrame,
        table: DataTable[str],
        show_index: bool = True,
        index_name: Optional[str] = None,
    ) -> DataTable[str]:
        """
        Convert a pandas.DataFrame obj into a rich.Table obj.

        Parameters
        ----------
        pandas_dataframe: DataFrame
            A Pandas DataFrame to be converted to a rich Table.
        table: DataTable[str]
            A DataTable that should be populated by the DataFrame values.
        show_index: bool
            Add a column with a row count to the table. Defaults to True.
        index_name: Optional[str]
            The column name to give to the index column. Defaults to None, showing no value.

        Returns
        -------
        DataTable[str]
            The DataTable instance passed, populated with the DataFrame values.
        """
        table.clear(columns=True)
        if show_index:
            index_name = str(index_name) if index_name else ""
            table.add_column(index_name)
        for column in pandas_dataframe.columns:
            table.add_column(str(column))
        pandas_dataframe.replace([np.NaN], [""], inplace=True)
        for index, value_list in enumerate(pandas_dataframe.values.tolist()):
            row = [str(index)] if show_index else []
            row += [str(x) for x in value_list]
            table.add_row(*row)
        return table


class FileSizeError(Exception):
    """
    File Too Large Error
    """


class CurrentFileInfoBar(Widget):
    """
    A Widget that displays information about the currently selected file

    Thanks, Kupo. https://github.com/darrenburns/kupo
    """

    file_info: Union[FileInfo, var[None]] = reactive(None)  # type: ignore[assignment]

    def watch_file_info(self, new_file: Union[FileInfo, None]) -> None:
        """
        Watch the file_info property for changes
        """
        if new_file is None:
            self.display = False
        else:
            self.display = True

    @classmethod
    def _convert_size(cls, size_bytes: int) -> str:
        """
        Convert Bytes to Human Readable String
        """
        if size_bytes == 0:
            return " 0B"
        size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
        index = int(math.floor(math.log(size_bytes, 1024)))
        p = math.pow(1024, index)
        number = round(size_bytes / p, 2)
        unit = size_name[index]
        return f"{number:.0f}{unit}"

    def render(self) -> RenderableType:
        """
        Render the Current File Info Bar
        """
        if self.file_info is None or not self.file_info.is_file:
            return Text("")
        status_string = "🗄️️️  " + self._convert_size(self.file_info.size)
        if self.file_info.last_modified is not None:
            modify_time = self.file_info.last_modified.strftime("%b %d, %Y %I:%M %p")
            status_string += "  📅  " + modify_time
        status_string += (
            "  💾  "
            + self.file_info.file.name
            + "  📂  "
            + self.file_info.file.parent.name
        )
        if self.file_info.owner not in ["", None]:
            status_string += "  👤  " + self.file_info.owner
        if self.file_info.group.strip() not in ["", None]:
            status_string += "  🏠  " + self.file_info.group
        return Text(status_string, style="dim")


class ConfirmationPopUp(Container):
    """
    A Pop Up that asks for confirmation
    """

    __confirmation_message__: str = dedent(
        """
        ## File Download

        Are you sure you want to download that file?
        """
    )

    def compose(self) -> ComposeResult:
        """
        Compose the Confirmation Pop Up
        """
        self.download_message = Static(Markdown(""))
        yield self.download_message
        yield Button("Yes", variant="success")
        yield Button("No", variant="error")

    @on(Button.Pressed)
    def handle_download_selection(self, message: Button.Pressed) -> None:
        """
        Handle Button Presses
        """
        self.app.confirmation_window.display = False  # type: ignore[attr-defined]
        if message.button.variant == "success":
            self.app.download_selected_file()  # type: ignore[attr-defined]
        self.app.table_view.display = self.app.hidden_table_view  # type: ignore[attr-defined]



"""
Screenshot Testing Using Cassettes!
"""

from textual_universal_directorytree import GitHubPath

from tests.conftest import cassette
from tests.helpers import Screenshotter


@cassette
def test_github_screenshot(github_release_path: GitHubPath) -> None:
    """
    Snapshot a release of this repo
    """
    file_path = str(github_release_path)
    screenshotter = Screenshotter(file_path=file_path)
    screenshot, screenshot_path = screenshotter.take_screenshot()
    assert screenshot_path.read_text() == screenshot


@cassette
def test_github_screenshot_license(github_release_path: GitHubPath) -> None:
    """
    Snapshot the LICENSE file
    """
    file_path = str(github_release_path / "LICENSE")
    screenshotter = Screenshotter(file_path=file_path)
    screenshot, screenshot_path = screenshotter.take_screenshot()
    assert screenshot_path.read_text() == screenshot



"""
Tools for browsr
"""

import pathlib
from typing import List, Optional

from textual._doc import take_svg_screenshot

from browsr import Browsr
from browsr._base import TextualAppContext


def take_screenshot(
    app: Optional[Browsr] = None,
    press: Optional[List[str]] = None,
    file_path: Optional[pathlib.Path] = None,
) -> str:
    """
    Take a screenshot of the app

    Parameters
    ----------
    app: Optional[Browsr]
        The browsr app, if None, a new one will be created
    press: Optional[List[str]]
        A list of keys to press
    file_path: Optional[pathlib.Path]
        The path to open in the app

    Returns
    -------
    str
        The SVG screenshot
    """
    if file_path is None:
        repo_dir = pathlib.Path(__file__).parent.parent.resolve()
        file_path = repo_dir
    if app is None:
        context = TextualAppContext(file_path=str(file_path))
        app = Browsr(config_object=context)
    screenshot = take_svg_screenshot(
        app=app,
        terminal_size=(160, 48),
        press=press or [],
        title=None,
    )
    return screenshot



"""
Testing of the Demo Class
"""

from click.testing import CliRunner

from browsr._cli import browsr


def test_cli_main(runner: CliRunner) -> None:
    """
    Test the main function of the CLI
    """
    result = runner.invoke(browsr, ["--help"])
    assert result.exit_code == 0



"""
Browsr TUI App

This module contains the code browser app for the browsr package.
This app was inspired by the CodeBrowser example from textual
"""

import json
import pathlib
import shutil
from os import getenv
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Iterable, Optional, Union

import pandas as pd
import upath
from art import text2art  # type: ignore[import]
from rich.markdown import Markdown
from rich.syntax import Syntax
from rich.traceback import Traceback
from rich_pixels import Pixels
from textual import on
from textual.binding import Binding
from textual.containers import Container, Horizontal, VerticalScroll
from textual.events import Mount
from textual.reactive import var
from textual.widget import Widget
from textual.widgets import DataTable, Footer, Header, Static
from textual_universal_directorytree import is_local_path, is_remote_path

from browsr._base import (
    BrowsrTextualApp,
    ConfirmationPopUp,
    CurrentFileInfoBar,
    FileSizeError,
    TextualAppContext,
)
from browsr._config import favorite_themes, image_file_extensions
from browsr._utils import (
    ArchiveFileError,
    FileInfo,
    get_file_info,
    handle_duplicate_filenames,
    open_image,
    render_file_to_string,
)
from browsr._version import __application__
from browsr.universal_directory_tree import (
    BrowsrDirectoryTree,
)


class Browsr(BrowsrTextualApp):
    """
    Textual code browser app.
    """

    TITLE = __application__
    CSS_PATH = "browsr.css"
    BINDINGS = [
        Binding("q", "quit", "Quit"),
        Binding("f", "toggle_files", "Toggle Files"),
        Binding("t", "theme", "Toggle Theme"),
        Binding("n", "linenos", "Toggle Line Numbers"),
        Binding("d", "toggle_dark", "Toggle Dark Mode"),
    ]

    show_tree = var(True)
    theme_index = var(0)
    linenos = var(False)
    rich_themes = favorite_themes
    selected_file_path: Union[upath.UPath, pathlib.Path, None, var[None]] = var(None)
    force_show_tree = var(False)
    hidden_table_view = var(False)

    def watch_show_tree(self, show_tree: bool) -> None:
        """
        Called when show_tree is modified.
        """
        self.set_class(show_tree, "-show-tree")

    def compose(self) -> Iterable[Widget]:
        """
        Compose our UI.
        """
        assert isinstance(self.config_object, TextualAppContext)
        file_path = self.config_object.path
        if is_remote_path(file_path):
            self.bind("x", "download_file", description="Download File", show=True)
        if file_path.is_file():
            self.selected_file_path = file_path
            file_path = file_path.parent
        elif file_path.is_dir() and file_path.joinpath("README.md").exists():
            if TYPE_CHECKING:
                assert isinstance(file_path, pathlib.Path)
            self.selected_file_path = file_path.joinpath("README.md")
            self.force_show_tree = True
        self.header = Header()
        yield self.header
        self.directory_tree = BrowsrDirectoryTree(str(file_path), id="tree-view")
        self.code_view = VerticalScroll(Static(id="code", expand=True), id="code-view")
        self.table_view: DataTable[str] = DataTable(
            zebra_stripes=True, show_header=True, show_cursor=True, id="table-view"
        )
        self.table_view.display = False
        self.confirmation = ConfirmationPopUp()
        self.confirmation_window = Container(
            self.confirmation, id="confirmation-container"
        )
        self.confirmation_window.display = False
        self.container = Container(
            self.directory_tree,
            self.code_view,
            self.table_view,
            self.confirmation_window,
        )
        yield self.container
        self.file_information = CurrentFileInfoBar()
        self.info_bar = Horizontal(
            self.file_information,
            id="file-info-bar",
        )
        if self.selected_file_path is not None:
            self.file_information.file_info = get_file_info(
                file_path=self.selected_file_path
            )
        yield self.info_bar
        self.footer = Footer()
        yield self.footer

    def render_document(
        self,
        file_info: FileInfo,
    ) -> Union[Syntax, Markdown, DataTable[str], Pixels]:
        """
        Render a Code Doc Given Its Extension

        Parameters
        ----------
        file_info: FileInfo
            The file info object for the file to render.

        Returns
        -------
        Union[Syntax, Markdown, DataTable[str], Pixels]
        """
        document = file_info.file
        if document.suffix == ".md":
            return Markdown(
                document.read_text(encoding="utf-8"),
                code_theme=self.rich_themes[self.theme_index],
                hyperlinks=True,
            )
        elif ".csv" in document.suffixes:
            df = pd.read_csv(document, nrows=1000)
            return self.df_to_table(pandas_dataframe=df, table=self.table_view)
        elif document.suffix == ".parquet":
            df = pd.read_parquet(document)[:1000]
            return self.df_to_table(pandas_dataframe=df, table=self.table_view)
        elif document.suffix.lower() in image_file_extensions:
            screen_width = self.app.size.width / 4
            content = open_image(document=document, screen_width=screen_width)
            return content
        elif document.suffix.lower() in [".json"]:
            code_str = render_file_to_string(file_info=file_info)
            try:
                code_obj = json.loads(code_str)
                code_lines = json.dumps(code_obj, indent=2).splitlines()
            except json.JSONDecodeError:
                code_lines = code_str.splitlines()
        else:
            code_str = render_file_to_string(file_info=file_info)
            code_lines = code_str.splitlines()
        code = "\n".join(code_lines[:1000])
        lexer = Syntax.guess_lexer(str(document), code=code)
        return Syntax(
            code=code,
            lexer=lexer,
            line_numbers=self.linenos,
            word_wrap=False,
            indent_guides=False,
            theme=self.rich_themes[self.theme_index],
        )

    def _handle_file_size(self, file_info: FileInfo) -> None:
        """
        Handle a File Size
        """
        file_size_mb = file_info.size / 1000 / 1000
        too_large = file_size_mb >= self.config_object.max_file_size  # type: ignore[union-attr]
        exception = (
            True
            if is_local_path(file_info.file) and ".csv" in file_info.file.suffixes
            else False
        )
        if too_large is True and exception is not True:
            raise FileSizeError("File too large")

    def _render_file(
        self, file_path: pathlib.Path, code_view: Static, font: str
    ) -> Union[Syntax, Markdown, DataTable[str], Pixels, None]:
        """
        Render a File
        """
        try:
            file_info = get_file_info(file_path=file_path)
            self._handle_file_size(file_info=file_info)
            element = self.render_document(file_info=file_info)
            return element
        except FileSizeError:
            self.table_view.display = False
            self.code_view.display = True
            code_view.update(
                text2art("FILE TOO", font=font) + "\n\n" + text2art("LARGE", font=font)
            )
            self.sub_title = f"ERROR [{self.rich_themes[self.theme_index]}]"
        except PermissionError:
            self.table_view.display = False
            self.code_view.display = True
            code_view.update(
                text2art("PERMISSION", font=font)
                + "\n\n"
                + text2art("ERROR", font=font)
            )
            self.sub_title = f"ERROR [{self.rich_themes[self.theme_index]}]"
        except UnicodeError:
            self.table_view.display = False
            self.code_view.display = True
            code_view.update(
                text2art("ENCODING", font=font) + "\n\n" + text2art("ERROR", font=font)
            )
            self.sub_title = f"ERROR [{self.rich_themes[self.theme_index]}]"
        except ArchiveFileError:
            self.table_view.display = False
            self.code_view.display = True
            code_view.update(
                text2art("ARCHIVE", font=font) + "\n\n" + text2art("FILE", font=font)
            )
            self.sub_title = f"ERROR [{self.rich_themes[self.theme_index]}]"
        except Exception:
            self.table_view.display = False
            self.code_view.display = True
            code_view.update(
                Traceback(theme=self.rich_themes[self.theme_index], width=None)
            )
            self.sub_title = "ERROR" + f" [{self.rich_themes[self.theme_index]}]"
        return None

    def render_code_page(
        self,
        file_path: pathlib.Path,
        scroll_home: bool = True,
        content: Optional[Any] = None,
    ) -> None:
        """
        Render the Code Page with Rich Syntax
        """
        code_view = self.query_one("#code", Static)
        font = "univers"
        if content is not None:
            code_view.update(text2art(content, font=font))
            return
        element = self._render_file(file_path=file_path, code_view=code_view, font=font)
        if isinstance(element, DataTable):
            self.code_view.display = False
            self.table_view.display = True
            if scroll_home is True:
                self.query_one(DataTable).scroll_home(animate=False)
        elif element is not None:
            self.table_view.display = False
            self.code_view.display = True
            code_view.update(element)
            if scroll_home is True:
                self.query_one("#code-view").scroll_home(animate=False)
            self.sub_title = f"{file_path} [{self.rich_themes[self.theme_index]}]"

    @on(Mount)
    def start_up_app(self) -> None:
        """
        On Application Mount - See If a File Should be Displayed
        """
        if self.selected_file_path is not None:
            self.show_tree = self.force_show_tree
            self.render_code_page(file_path=self.selected_file_path)
        else:
            self.show_tree = True
            self.render_code_page(
                file_path=pathlib.Path.cwd(), content=__application__.upper()
            )

    @on(BrowsrDirectoryTree.FileSelected)
    def handle_file_selected(self, message: BrowsrDirectoryTree.FileSelected) -> None:
        """
        Called when the user click a file in the directory tree.
        """
        self.selected_file_path = upath.UPath(message.path)
        file_info = get_file_info(file_path=self.selected_file_path)
        self.render_code_page(file_path=upath.UPath(message.path))
        self.file_information.file_info = file_info

    def action_toggle_files(self) -> None:
        """
        Called in response to key binding.
        """
        self.show_tree = not self.show_tree

    def action_theme(self) -> None:
        """
        An action to toggle rich theme.
        """
        if self.selected_file_path is None:
            return
        elif self.theme_index < len(self.rich_themes) - 1:
            self.theme_index += 1
        else:
            self.theme_index = 0
        self.render_code_page(file_path=self.selected_file_path, scroll_home=False)

    def action_linenos(self) -> None:
        """
        An action to toggle line numbers.
        """
        if self.selected_file_path is None:
            return
        self.linenos = not self.linenos
        self.render_code_page(file_path=self.selected_file_path, scroll_home=False)

    def _get_download_file_name(self) -> pathlib.Path:
        """
        Get the download file name.
        """
        download_dir = pathlib.Path.home() / "Downloads"
        if not download_dir.exists():
            raise FileNotFoundError(f"Download directory {download_dir} not found")
        download_path = download_dir / self.selected_file_path.name  # type: ignore[union-attr]
        handled_download_path = handle_duplicate_filenames(file_path=download_path)
        return handled_download_path

    def download_selected_file(self) -> None:
        """
        Download the selected file.
        """
        if self.selected_file_path is None:
            return
        elif self.selected_file_path.is_dir():
            return
        elif is_remote_path(self.selected_file_path):
            handled_download_path = self._get_download_file_name()
            with self.selected_file_path.open("rb") as file_handle:
                with handled_download_path.open("wb") as download_handle:
                    shutil.copyfileobj(file_handle, download_handle)

    def action_download_file(self) -> None:
        """
        Download the selected file.
        """
        if self.selected_file_path is None:
            return
        elif self.selected_file_path.is_dir():
            return
        elif is_remote_path(self.selected_file_path):
            handled_download_path = self._get_download_file_name()
            prompt_message: str = dedent(
                f"""
                ## File Download

                **Are you sure you want to download that file?**

                **File:** `{self.selected_file_path}`

                **Path:** `{handled_download_path}`
                """
            )
            self.confirmation.download_message.update(Markdown(prompt_message))
            self.confirmation.refresh()
            self.hidden_table_view = self.table_view.display
            self.table_view.display = False
            self.confirmation_window.display = True


app = Browsr(
    config_object=TextualAppContext(file_path=getenv("BROWSR_PATH"), debug=True)
)

if __name__ == "__main__":
    app.run()



"""
Helpers for the tests
"""

import pathlib
from os import environ, getenv
from typing import List, Optional, Tuple

from browsr import Browsr
from browsr._base import TextualAppContext
from browsr._tools import take_screenshot


class Screenshotter:
    """
    App Screenshotter
    """

    def __init__(self, file_path: str) -> None:
        """
        Initialize the Screenshotter
        """
        self.screenshot_dir = file_path
        self.context = TextualAppContext(
            file_path=self.screenshot_dir,
        )
        self.app = Browsr(config_object=self.context)

    def take_screenshot(
        self, press: Optional[List[str]] = None
    ) -> Tuple[str, pathlib.Path]:
        """
        Take a Screenshot
        """
        screenshot = take_screenshot(app=self.app, press=press)
        screenshot_path = self._get_screenshot_path()
        if getenv("BROWSR_REGENERATE_SCREENSHOTS", "0") != "0":
            screenshot_path.write_text(screenshot)
        return screenshot, screenshot_path

    @classmethod
    def _get_screenshot_path(cls) -> pathlib.Path:
        """
        Get the Screenshot Path
        """
        test_dir = pathlib.Path(__file__).parent
        screenshot_dir = test_dir / "screenshots"
        current_test = environ["PYTEST_CURRENT_TEST"].split("::")[-1].split(" ")[0]
        screenshot_path = screenshot_dir / f"{current_test}.svg"
        return screenshot_path



"""
A universal directory tree widget for Textual.
"""

from __future__ import annotations

from typing import Iterable, Optional

from textual.widgets._directory_tree import DirEntry
from textual.widgets._tree import TreeNode
from textual_universal_directorytree import UniversalDirectoryTree
from upath import UPath as Path


class BrowsrDirectoryTree(UniversalDirectoryTree):
    """
    A DirectoryTree that can handle any filesystem.
    """

    @classmethod
    def _handle_top_level_bucket(cls, dir_path: Path) -> Optional[Iterable[Path]]:
        """
        Handle scenarios when someone wants to browse all of s3

        This is because S3FS handles the root directory differently than other filesystems
        """
        if str(dir_path) == "s3:/":
            sub_buckets = sorted(
                Path(f"s3://{bucket.name}") for bucket in dir_path.iterdir()
            )
            return sub_buckets
        return None

    def _populate_node(self, node: TreeNode[DirEntry], content: Iterable[Path]) -> None:  # type: ignore[override]
        """
        Populate the given tree node with the given directory content.

        This function overrides the original textual method to handle root level
        cloud buckets.
        """
        top_level_buckets = self._handle_top_level_bucket(dir_path=node.data.path)  # type: ignore[union-attr, arg-type]
        if top_level_buckets is not None:
            content = top_level_buckets
        node.remove_children()
        for path in content:
            if top_level_buckets is not None:
                path_name = str(path).replace("s3://", "").rstrip("/")
            else:
                path_name = path.name
            node.add(
                path_name,
                data=DirEntry(path),
                allow_expand=self._safe_is_dir(path),
            )
        node.expand()
